/*#define GRBS_ROUTE_ARC_TRACE*/

#undef tprinf
#ifdef GRBS_ROUTE_ARC_TRACE
#include <stdio.h>
#define tprintf printf
#else
#define tprintf grbs_nullprintf
#endif

#include "geo_isc.h"
#include <gengeo2d/carc.h>
#include <gengeo2d/sline.h>
#include <gengeo2d/intersect.h>

static void arc_bbox(grbs_arc_t *a)
{
	g2d_box_t bb;
	g2d_carc_t carc;

	carc.c.x = a->parent_pt->x;
	carc.c.y = a->parent_pt->y;
	carc.r = a->r + a->copper + a->clearance; /* the outmost radius, clearance included */
	carc.start = a->sa; carc.delta = a->da;

	bb = g2d_carc_bbox(&carc);
	a->bbox.x1 = bb.p1.x; a->bbox.y1 = bb.p1.y;
	a->bbox.x2 = bb.p2.x; a->bbox.y2 = bb.p2.y;
}

static void arc_reg(grbs_t *grbs, grbs_arc_t *a)
{
	grbs_rtree_insert(&grbs->arc_tree, a, &a->bbox);
}

static void arc_unreg(grbs_t *grbs, grbs_line_t *a)
{
	grbs_rtree_delete(&grbs->arc_tree, a, &a->bbox);
}

static int grbs_isc_carc_linecop(g2d_carc_t *carc, grbs_line_t *line, grbs_2net_t *ltn)
{
	g2d_sline_t sline;
	g2d_cline_t cp, cn;
	double lcop = ltn->copper;
	g2d_vect_t ip[2];


	sline.c.p1.x = line->x1; sline.c.p1.y = line->y1;
	sline.c.p2.x = line->x2; sline.c.p2.y = line->y2;
	sline.s.width = lcop + lcop - 0.001; /* leave a margin of 1 um so the test is exclusive */
	sline.s.cap = G2D_CAP_ROUND;

	g2d_sline_sides(&sline, &cp, &cn);
	if (g2d_iscp_cline_carc(&cp, carc,  ip, NULL, 0)) {
		tprintf("ISC: sline carc 1: %f %f\n", ip[0].x, ip[0].y);
		return 1;
	}
	if (g2d_iscp_cline_carc(&cn, carc,  ip, NULL, 0)) {
		tprintf("ISC: sline carc 2: %f %f\n", ip[0].x, ip[0].y);
		return 1;
	}
	return 0;
}

static int grbs_isc_line_carccop(g2d_carc_t *carc, grbs_line_t *line, grbs_2net_t *ltn)
{
	g2d_sline_t sline;
	g2d_cline_t cp, cn;
	double lcop = ltn->copper, lclr = ltn->clearance;
	g2d_vect_t ip[2];


	sline.c.p1.x = line->x1; sline.c.p1.y = line->y1;
	sline.c.p2.x = line->x2; sline.c.p2.y = line->y2;
	sline.s.width = lcop + lcop + lclr + lclr - 0.001; /* leave a margin of 1 um so the test is exclusive */
	sline.s.cap = G2D_CAP_ROUND;

	g2d_sline_sides(&sline, &cp, &cn);
	if (g2d_iscp_cline_carc(&cp, carc,  ip, NULL, 0)) {
		tprintf("ISC: sline carc 3: %f %f\n", ip[0].x, ip[0].y);
		return 1;
	}
	if (g2d_iscp_cline_carc(&cn, carc,  ip, NULL, 0)) {
		tprintf("ISC: sline carc 4: %f %f\n", ip[0].x, ip[0].y);
		return 1;
	}
	return 0;
}

static grbs_line_t *arc_line_collision(grbs_t *grbs, grbs_2net_t *tn, grbs_arc_t *arc, int new)
{
	g2d_carc_t carc;
	double cop = tn->copper, clr = tn->clearance, origr;
	g2d_box_t bb;
	grbs_line_t *l;
	grbs_rtree_it_t it;

	carc.c.x = arc->parent_pt->x;
	carc.c.y = arc->parent_pt->y;

	if (new) {
		carc.r = origr = arc->new_r + cop + clr; /* the outmost radius, clearance included */
		carc.start = arc->new_sa;
		carc.delta = arc->new_da;
		bb = g2d_carc_bbox(&carc);
	}
	else {
		carc.r = origr = arc->r + cop + clr; /* the outmost radius, clearance included */
		carc.start = arc->sa;
		carc.delta = arc->da;
		bb.p1.x = arc->bbox.x1; bb.p1.y = arc->bbox.y1;
		bb.p2.x = arc->bbox.x2; bb.p2.y = arc->bbox.y2;
	}

	/* check if the outer clearance arc cuts into any line copper */
	for(l = grbs_rtree_first(&it, &grbs->line_tree, &bb); l != NULL; l = grbs_rtree_next(&it)) {
		grbs_2net_t *ltn = grbs_arc_parent_2net(l->a1);

		if (tn == ltn) continue; /* self-intersection, e.g. line end meets arc */

		/* lines starting or ending around our point are handled by arc angles;
		   they may temporarily intersect while bumping segments up */
		if ((l->a1->parent_pt == arc->parent_pt) || (l->a2->parent_pt == arc->parent_pt))
			continue;

		/* check if any arc clearance cuts into the line's copper */
		if (grbs_isc_carc_linecop(&carc, l, ltn))
			return l;

	/* check if any line outer clearance cline cuts arc cuts into the arc */
		carc.r -= clr;
		if (grbs_isc_line_carccop(&carc, l, ltn))
			return l;
		carc.r = origr;
	}

	return NULL;
}

static int grbs_isc_carc_carc(g2d_carc_t *c1, g2d_carc_t *c2, int dbg)
{
	g2d_vect_t ip[2];

	if (g2d_iscp_carc_carc(c1, c2, ip, NULL)) {
		tprintf("ISC: carc carc %d: %f %f\n", dbg, ip[0].x, ip[0].y);
		return 1;
	}
	return 0;
}

static grbs_arc_t *arc_arc_collision(grbs_t *grbs, grbs_2net_t *tn, grbs_arc_t *arc, int new)
{
	g2d_carc_t carc, ca;
	double cop = tn->copper, clr = tn->clearance, origr;
	g2d_box_t bb;
	grbs_arc_t *a;
	grbs_rtree_it_t it;

	carc.c.x = arc->parent_pt->x;
	carc.c.y = arc->parent_pt->y;

	if (new) {
		carc.r = origr = arc->new_r + cop + clr; /* the outmost radius, clearance included */
		carc.start = arc->new_sa;
		carc.delta = arc->new_da;
		bb = g2d_carc_bbox(&carc);
	}
	else {
		carc.r = origr = arc->r + cop + clr; /* the outmost radius, clearance included */
		carc.start = arc->sa;
		carc.delta = arc->da;
		bb.p1.x = arc->bbox.x1; bb.p1.y = arc->bbox.y1;
		bb.p2.x = arc->bbox.x2; bb.p2.y = arc->bbox.y2;
	}

	/* check if the outer clearance arc cuts into any line copper */
	for(a = grbs_rtree_first(&it, &grbs->arc_tree, &bb); a != NULL; a = grbs_rtree_next(&it)) {
		grbs_2net_t *atn = grbs_arc_parent_2net(a);

		if (tn == atn) continue; /* self-intersection, e.g. line end meets arc */

		/* arcs around our point are handled by angles; they may temporarily
		   intersect while bumping segments up */
		if (a->parent_pt == arc->parent_pt)
			continue;

		ca.c.x = a->parent_pt->x; ca.c.y = a->parent_pt->y;
		ca.start = a->sa; ca.delta = a->da;

		/* check if our arc clearance cuts into the a's copper */
		ca.r = a->r + atn->copper;
		if (grbs_isc_carc_carc(&carc, &ca, 1))
			return a;

		/* check if our arc copper cuts into the a's clearance */
		carc.r -= clr;
		ca.r += atn->clearance;
		if (grbs_isc_carc_carc(&carc, &ca, 2))
			return a;
		carc.r = origr;
	}

	return NULL;
}

static grbs_point_t *arc_point_collision(grbs_t *grbs, grbs_2net_t *tn, grbs_arc_t *arc, int new)
{
	g2d_carc_t carc, ca;
	double cop = tn->copper, clr = tn->clearance, origr;
	g2d_box_t bb;
	grbs_point_t *p;
	grbs_rtree_it_t it;

	carc.c.x = arc->parent_pt->x;
	carc.c.y = arc->parent_pt->y;

	if (new) {
		if (arc->new_r == 0)
			return 0; /* incident arc won't ever collide */
		carc.r = origr = arc->new_r + cop + clr; /* the outmost radius, clearance included */
		carc.start = arc->new_sa;
		carc.delta = arc->new_da;
		bb = g2d_carc_bbox(&carc);
	}
	else {
		if (arc->r == 0)
			return 0; /* incident arc won't ever collide */
		carc.r = origr = arc->r + cop + clr; /* the outmost radius, clearance included */
		carc.start = arc->sa;
		carc.delta = arc->da;
		bb.p1.x = arc->bbox.x1; bb.p1.y = arc->bbox.y1;
		bb.p2.x = arc->bbox.x2; bb.p2.y = arc->bbox.y2;
	}

	/* check if arc cuts into any point */
	for(p = grbs_rtree_first(&it, &grbs->point_tree, &bb); p != NULL; p = grbs_rtree_next(&it)) {

		/* arcs around our point are handled by angles; they may temporarily
		   intersect while bumping segments up */
		if (arc->parent_pt == p)
			continue;

		ca.c.x = p->x; ca.c.y = p->y;
		ca.start = 0; ca.delta = 2*M_PI;

		/* check if our arc clearance cuts into the p's copper */
		ca.r = p->copper;
		if (grbs_isc_carc_carc(&carc, &ca, 3))
			return p;

		/* check if our arc copper cuts into the a's clearance */
		carc.r -= clr;
		ca.r += p->clearance;
		if (grbs_isc_carc_carc(&carc, &ca, 4))
			return p;
		carc.r = origr;
	}

	return NULL;
}
